"""
Loads the opavote schema
"""

from rcvformats.schemas.base import GenericJsonSchema


#  pylint:disable=invalid-name
class SchemaV1_0(GenericJsonSchema):
    """ Schema for the initial Opavote schema """

    @property
    def schema_filename(self):
        return 'opavote.schema.json'

    def version(self):
        return "1.0"

#  pylint:disable=invalid-name


class SchemaV1_1(GenericJsonSchema):
    """ Schema for the version created on 2022-05-18 """

    @property
    def schema_filename(self):
        return 'opavote_v1_1.schema.json'

    def version(self):
        return "1.1"



"""
Attempts to convert any file to the standard format.
If you know the file format, you should not use this - it will
loop through all schemas which will be needlessly slow.
"""

import json

from rcvformats.conversions.base import CouldNotConvertException
from rcvformats.conversions.dominion import DominionConverter
from rcvformats.conversions.dominion_first_round_only import DominionFirstRoundOnlyConverter
from rcvformats.conversions.electionbuddy import ElectionBuddyConverter
from rcvformats.conversions.opavote import OpavoteConverter
from rcvformats.conversions.ut_without_transfers import UTWithoutTransfersConverter
from rcvformats.conversions.base import Converter


class AutomaticConverter(Converter):
    """ Interface for converters """

    def __init__(self):
        # In order of likelihood of a hit - just my guess.
        # As this list grows, we should really consider doing something intelligent.
        self.converters = [
            DominionConverter,
            ElectionBuddyConverter,
            OpavoteConverter,
            DominionFirstRoundOnlyConverter
        ]

        super().__init__()

    def _convert_file_object_to_ut(self, file_object):
        # If it matches the schema already, load and return the data
        if self.ut_schema.validate(file_object):
            file_object.seek(0)
            return json.load(file_object)

        # Otherwise, try each converter - skipping schemas for speed
        additional_errors = []
        for converterType in self.converters:
            file_object.seek(0)
            try:
                data = converterType().convert_to_ut(file_object)
                converter = UTWithoutTransfersConverter(allow_guessing=False)
                return converter.fill_in_tally_data(data)
            except CouldNotConvertException as exception:
                additional_errors.append(converterType.__name__ + ":" + str(exception))
                continue

        # If it failed, accumulate all errors from schemas
        error_message = "When trying to parse as the Universal Tabulator schema, " +\
            f"received error: \"{str(self.ut_schema.last_error())}\". " +\
            "Further, it did not match any other known format. Additional errors: \n\n" +\
            '\n\n'.join(additional_errors)
        raise CouldNotConvertException(error_message)



"""
Reads an Dominion XLSX results file, writes to the standard format
"""

import math
import xml.etree.ElementTree as ET

from rcvformats.conversions.base import Converter


class DominionFirstRoundOnlyConverter(Converter):
    """
    Parses the dominion first-round-only file format as exemplified in
    /testdata/inputs/dominionFirstRoundOnly. These are .xml files.
    """

    def _convert_file_object_to_ut(self, file_object):
        element_tree = ET.parse(file_object)
        contest_id_element = element_tree.getroot() \
            .find('{ElectionSummaryReportRPT}tabBatchIdList') \
            .find('{ElectionSummaryReportRPT}TabBatchGroup_Collection') \
            .find('{ElectionSummaryReportRPT}TabBatchGroup') \
            .find('{ElectionSummaryReportRPT}ElectionSummarySubReport') \
            .find('{ElectionSummaryReportRPT}Report') \
            .find('{ElectionSummaryReportRPT}contestList') \
            .find('{ElectionSummaryReportRPT}ContestIdGroup_Collection') \
            .find('{ElectionSummaryReportRPT}ContestIdGroup')

        config = self._parse_config(element_tree, contest_id_element)
        results = self._parse_vote_count(contest_id_element)
        config['threshold'] = self._threshold_from(results)
        urcvt_data = {'config': config, 'results': results}

        return urcvt_data

    @classmethod
    def _parse_config(cls, element_tree, contest_id_element):
        config = {}
        config['date'] = element_tree.getroot() \
            .find('{ElectionSummaryReportRPT}Title') \
            .find('{ElectionSummaryReportRPT}Report') \
            .get('Textbox9')
        # Remove the timestamp from the date
        config['date'] = config['date'][:10]
        config['contest'] = contest_id_element.get('contestId')

        return config

    @classmethod
    def _parse_vote_count(cls, contest_id_element):
        tally = {}
        candidates = contest_id_element \
            .find('{ElectionSummaryReportRPT}CandidateResults') \
            .find('{ElectionSummaryReportRPT}Report') \
            .find('{ElectionSummaryReportRPT}Tablix1') \
            .find('{ElectionSummaryReportRPT}chGroup_Collection')
        for candidate in candidates.findall('{ElectionSummaryReportRPT}chGroup'):
            candidate_data = candidate.find('{ElectionSummaryReportRPT}candidateNameTextBox4')
            name = candidate_data.get('candidateNameTextBox4')
            num_votes = candidate_data.find('{ElectionSummaryReportRPT}Textbox13').get('vot8')
            tally[name] = float(num_votes)

        return [{
            "round": 1,
            "tally": tally,
            "tallyResults": []
        }]

    @classmethod
    def _threshold_from(cls, results):
        total = sum(results[0]['tally'].values())
        return math.floor(total / 2 + 1)



#!/usr/bin/env python

"""
A command-line interface to some of the python functionality
"""

import argparse
import json
import sys

from enum import Enum

from rcvformats.schemas import electionbuddy
from rcvformats.schemas import universaltabulator
from rcvformats.schemas import opavote
from rcvformats.conversions.automatic import AutomaticConverter
from rcvformats.conversions.ut_without_transfers import UTWithoutTransfersConverter


class FormatEnum(Enum):
    """ Enum for validator, and perhaps eventually converter. """
    UT = 'ut'
    EB = 'eb'
    OV10 = 'ov10'
    OV11 = 'ov11'

    def __str__(self):
        return str(self.value)


def convert(input_filename, output_filename):
    """ Automatic converter from input_filename to output_filename """
    standardized_format = AutomaticConverter().convert_to_ut(input_filename)
    with open(output_filename, 'w', encoding='utf-8') as file_obj:
        json.dump(standardized_format, file_obj)


def validate(input_filename, schema):
    """ validates input_filename with schema """
    if schema == FormatEnum.UT:
        schema = universaltabulator.SchemaV0()
    elif schema == FormatEnum.EB:
        schema = electionbuddy.SchemaV0()
    elif schema == FormatEnum.OV10:
        schema = opavote.SchemaV1_0()
    elif schema == FormatEnum.OV11:
        schema = opavote.SchemaV1_1()

    is_valid = schema.validate(input_filename)
    if is_valid:
        print("Schema is valid.")
    else:
        print("Schema is not valid. Errors: ", schema.last_error())


def add_transfers(input_filename, output_filename, allow_guessing):
    """ Adds tally transfers if they don't exist. Overwrites them if they do. """
    # Adding transfers, internally, is just another conversion
    converter = UTWithoutTransfersConverter(allow_guessing=allow_guessing)
    with_transfers = converter.convert_to_ut(input_filename)
    with open(output_filename, 'w', encoding='utf-8') as file_obj:
        json.dump(with_transfers, file_obj)


def _add_input_arg(parser):
    parser.add_argument(
        '-i',
        '--input',
        dest='input_filename',
        help='File to convert',
        required=True)


def _add_output_arg(parser):
    parser.add_argument(
        '-o',
        '--output',
        dest='output_filename',
        help='Where to place the JSON file on success',
        required=True)


def main():
    """ Main function: cli entrypoint, using argparse """
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(
        dest='subparser',
        help='Tooling for RCV Formats. See more at https://rcvformats.readthedocs.io/en/latest/')

    conv_parser = subparsers.add_parser(
        'convert', help='Converts from whatever format you have to the Universal Tabulator format.')
    _add_input_arg(conv_parser)
    _add_output_arg(conv_parser)

    validate_parser = subparsers.add_parser(
        'validate', help='Validates the file with one of the three accepted formats')
    _add_input_arg(validate_parser)
    validate_parser.add_argument(
        '-s',
        '--schema',
        dest='schema',
        type=FormatEnum,
        choices=list(FormatEnum),
        help='Schema to validate against: ut (Universal Tabulator/RCTab), '
             'eb (ElectionBuddy), or ov (OpaVote)',
        required=True)

    xfer_parser = subparsers.add_parser(
        'transfer', help='Adds "transfers" to the tallyResults of an otherwise-valid UI format')
    _add_input_arg(xfer_parser)
    _add_output_arg(xfer_parser)
    xfer_parser.add_argument(
        '-g',
        '--allow-guessing',
        dest='allow_guessing',
        action='store_true',
        help='During batch elimination, may we guess at the vote transfers? '
             'If not, will leave transfers blank for all batch elimination rounds.',
        required=False)

    args = parser.parse_args()
    if args.subparser is None:
        print(parser.print_help())
        sys.exit(-1)

    if args.subparser == 'convert':
        convert(args.input_filename, args.output_filename)
    if args.subparser == 'validate':
        validate(args.input_filename, args.schema)
    if args.subparser == 'transfer':
        add_transfers(args.input_filename, args.output_filename, args.allow_guessing)
